﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Text;
using Falcon.Extend;
using System.Linq;
using System.Threading;

namespace Falcon.TaskScheduling
{
    /// <summary>
    /// 安全队列实现
    /// </summary>
    public class ConcurrentQueueTask:ITaskScheduler
    {
        /// <summary>
        /// 选项
        /// </summary>
        public TaskSchedulerOption Option { get; set; }
        /// <summary>
        /// 系统级队列
        /// </summary>
        public ConcurrentQueue<ITaskUnit> SystemQueue;
        /// <summary>
        /// 优先级队列
        /// </summary>
        public ConcurrentQueue<ITaskUnit> FirstQueue;
        /// <summary>
        /// 空闲级队列
        /// </summary>
        public ConcurrentQueue<ITaskUnit> IdleQueue;
        /// <summary>
        /// 任务队列
        /// </summary>
        public List<QueueUnit> Queue;

        /// <summary>
        /// 通过提供的选项实例化任务调度程序
        /// </summary>
        /// <param name="option">配置参数</param>
        public ConcurrentQueueTask(TaskSchedulerOption option = null) {
            this.Option = option ?? new TaskSchedulerOption();
            if(this.Option.EnableSystemQueue) {
                this.SystemQueue = GetNewQueue();
            }
            if(this.Option.EnableFirstQueue) {
                this.FirstQueue = GetNewQueue();
            }
            if(this.Option.EnableIdleQueue) {
                this.IdleQueue = GetNewQueue();
            }
            this.Queue = new List<QueueUnit>();
            if(this.Option.QueueDefine.IsNotNullOrEmpty()) {
                var spc = ",;，；";
                var q = this.Option.QueueDefine.Split(spc.ToCharArray());
                foreach(var i in q) {
                    if(int.TryParse(i,out var number)) {
                        this.Queue.Add(new QueueUnit {
                            Queue = GetNewQueue(),
                            Weight = number,
                        });
                    } else {
                        throw new Exception($"QueueDefine值{this.Option.QueueDefine}定义错误");
                    }
                }
            }
        }
        /// <summary>
        /// 获取一个新的安全队列
        /// </summary>
        protected ConcurrentQueue<ITaskUnit> GetNewQueue() {
            return new ConcurrentQueue<ITaskUnit>();
        }

        public void AddFirstQueue(ITaskUnit task) {
            if(!this.Option.EnableFirstQueue) {
                throw new NotSupportedException();
            }
            this.FirstQueue.Enqueue(task);
        }

        public void AddIdleQueue(ITaskUnit task) {
            if(!this.Option.EnableIdleQueue) {
                throw new NotSupportedException();
            }
            this.IdleQueue.Enqueue(task);
        }

        public void AddSystemQueue(ITaskUnit task) {
            if(!this.Option.EnableSystemQueue) {
                throw new NotSupportedException();
            }
            this.SystemQueue.Enqueue(task);
        }

        public void AddQueue(int weight,ITaskUnit task) {
            var c = this.Queue.Count();
            if(c == 0) {
                throw new NotSupportedException();
            }
            var queue = this.Queue.Where(m => m.Weight == weight);
            if(!queue.Any()) {
                throw new Exception("添加到任务队列提供的weight值错误。必须提供TaskSchedulerOption.QueueDefine中定义的值。");
            }
            queue.First().Queue.Enqueue(task);
        }

        private bool _isRun = false;
        private static object _lockObj = new object();
        /// <summary>
        /// 开始执行调度任务
        /// </summary>
        public void Start() {
            if(!_isRun) {
                lock(_lockObj) {
                    if(!_isRun) {
                        _isRun = true;
                        var task = GetNextTask();
                        while(task != null) {
                            object result = null;
                            Exception exp = null;
                            try {
                                result = task.Run();
                            } catch(Exception ex) {
                                exp = ex;
                            }
                            task.TaskCallback(result,exp);
                            Thread.Sleep(0);
                            task = GetNextTask();
                        }
                        _isRun = false;
                    }
                }
            }
        }

        protected ITaskUnit GetNextTask() {
            ITaskUnit t;
            if(this.Option.EnableSystemQueue) {
                if(this.SystemQueue.TryDequeue(out t)) {
                    return t;
                }
            }
            if(this.Option.EnableFirstQueue) {
                if(this.FirstQueue.TryDequeue(out t)) {
                    return t;
                }
            }
            t = GetTaskFromQueue();
            if(t != null) {
                return t;
            }
            if(this.Option.EnableIdleQueue) {
                if(this.IdleQueue.TryDequeue(out t)) {
                    return t;
                }
            }
            return null;
        }

        private ITaskUnit GetTaskFromQueue() {
            var ql = this.Queue.Where(m => m.HasTask);
            var weight = ql.Sum(m => m.Weight);
            var point = GetRander(weight);
            foreach(var q in ql) {
                if(point < q.Weight) {
                    if(q.Queue.TryDequeue(out var tu)) {
                        return tu;
                    }
                }
                point -= q.Weight;
            }
            return null;
        }

        private static Random RandomNum = new Random();

        protected int GetRander(int max) {
            return RandomNum.Next(0,max);
        }
    }
}
